#ifndef ERA_COMMANDS_H
#define ERA_COMMANDS_H

#include "base/application.h"

//----------------------------------------------------------------

namespace era {
	extern base::command era_check_cmd;
	extern base::command era_dump_cmd;
	extern base::command era_invalidate_cmd;
	extern base::command era_restore_cmd;
}

//----------------------------------------------------------------

#endif
